//	xmmexposure.cc --- implementation of the XmmExposure class
//
//	DESCRIPTION:
//
//	AUTHOR:
//		D. Hinshaw
//

#include <cstdio>
#include <unistd.h>
#include <iostream>
#include <sstream>

#include "Xmm.h"

#include "param.h"
#include "errstr.h"
#include "Dal.h"
#include "DataSubspace.h"
#include "DataFilter.h"
#include "xmmexposure.h"
#include "OGIPgti.h"

XmmExposure::XmmExposure(Table *eventTable) : _eventTable(eventTable) {

  _set = _eventTable->parent();
  _instrument = Xmm::INSTRUMENT_NONE;

  // Do some checks to make sure we can do exposure correction on this data set

  if( !_set->hasAttribute("TELESCOP") ||
      _set->attribute("TELESCOP")->asString() != "XMM" ){
    errstr << "Can only update exposure information on XMM data." << Sas::warning("NotXMM");
    return;
  }

  _instrument = Xmm::toInstrument(_set->attribute("INSTRUME")->asString());

  // Get the list of CCDs with exposure extensions

  if( !_eventTable->hasColumn("CCDNR") ){
      errstr << "Event list does not have a CCDNR column. "
	     << "Cannot update exposure information. "
	     << Sas::warning("NoCCDcolumn");
      return;    
  }

  Column *ccdnr = _eventTable->column("CCDNR");
  ColumnData * cCcdnrDat = ccdnr->data();
  int8 * ccdnrDat = cCcdnrDat->int8Data();

  for(int i=1; i<=17; i++){
    std::stringstream extName;
    extName << "EXPOSU";
    extName.width(2);
    extName.fill('0');
    extName << i;
    if( !_set->hasBlock(extName.str()) ) continue;

    bool hasCcd=false;
    for(unsigned irow=0; irow<_eventTable->rows(); irow++){
      if(i==ccdnrDat[irow]){
	hasCcd=true;
	break;
      }
    }

    if(hasCcd)
      _ccdNums.push_back(i);
    else
      _set->deleteBlock(extName.str());

    
    /*
     * AI: Check if there is a PHASE keyword in the expression. If so, check that 
     * the user has run phasecalc for the EXPOSU table, if not raise a warning message.
     */
    DataSubspace dss(_eventTable);
    unsigned ncomps = dss.components();
    for(unsigned icomp=0; icomp<ncomps; icomp++)
      {
	DScomp * comp = dss.component(icomp);
	if( comp->hasFilter("PHASE", DSS::RangeT) )
	  { 
	    if( _set->hasBlock(extName.str()) )	    
	      if (!_set->table(extName.str())->hasColumn("PHASE") )
		errstr << "You have used the PHASE keyword to filter the event list.  "
		       << "You have to run phasecalc task for all the EXPOSUnn tables"
		       << Sas::error("PhaseTimeFilter");
	  }
      }
  }

  return;
}


void XmmExposure::getExposureInformation() {

  if( _instrument==Xmm::EMOS1 || _instrument==Xmm::EMOS2 ||
      _instrument==Xmm::EPN ){
    if(_instrument==Xmm::EPN) _targetCCD = 4;
    else _targetCCD = 1;
    _getGtiLists();
  }else if( _instrument==Xmm::RGS1 || _instrument==Xmm::RGS2 ){

    // Check for any time based filters

    DataSubspace dss(_eventTable);
    unsigned ncomps = dss.components();

    for(unsigned icomp=0; icomp<ncomps; icomp++){
      DScomp * comp = dss.component(icomp);
      if( comp->hasFilter("TIME", DSS::RangeT) || comp->hasFilter("TIME", DSS::RangeBlockT) ||
	  comp->hasFilter("TIME", DSS::GtiT) ){
	errstr << "You have used the TIME column to filter an RGS events list.  "
	       << "This will invalidate any exposure information in the produced products."
	       << Sas::warning("RgsTimeFilter");
	return;
      }
    }
  }
  
  return;
}


void XmmExposure::updateOntime() {

  if(_instrument==Xmm::INSTRUMENT_NONE ||
	 _instrument==Xmm::RGS1 ||
	 _instrument==Xmm::RGS2) return;

  // Sum up all the GTIs to get the ontime    

  for(unsigned iccd=0; iccd<_ccdNums.size(); iccd++){
    double ontime = 0;
    assert(_gtiLists.size() > iccd);
    for(unsigned igti=0; igti<_gtiLists[iccd].size(); igti++)
      ontime += _gtiLists[iccd][igti].stop() - _gtiLists[iccd][igti].start();

    // Write the new ontime keywords.

    std::stringstream keyName;
    keyName << "ONTIME";
    keyName.width(2);
    keyName.fill('0');
    keyName << _ccdNums[iccd];
    
    *_eventTable->addAttribute(keyName.str(), "Sum of GTIs for a particular CCD") 
      = ontime;
  }

  // Write ONTIME keyword for target CCD

  std::stringstream keyName;
  keyName << "ONTIME";
  keyName.width(2);
  keyName.fill('0');
  keyName << _targetCCD;

  if( _eventTable->hasAttribute(keyName.str()) )
    *_eventTable->addAttribute("ONTIME", "Sum of GTIs for central CCD") 
      = _eventTable->attribute(keyName.str())->asReal();
  else if( _ccdNums.size()>0 ){
    errstr << "Ontime for target "
	   << "CCD (number " << _targetCCD << ") is not available. "
	   << "Setting ONTIME to zero."
	   << Sas::warning("NoTargetOntime");
    *_eventTable->addAttribute("ONTIME", "Sum of GTIs for central CCD") = 0;
  }

	// AI: SOC-SPR-2629:
	// If there are not events as a result of the filtering, the _ccdNums vector 
	// (the vector with the ccd numbers that contains events) is empty and therefore
	// the _gtiList is also empty, so the updateOntime method does not uptdate the ONTIME
	// keyword. 
	// To fix this problem, if the _ccdNums vector is empty, we check for all the 
	// ONTIME keywords in the header file and they are updated with 0 time. 
	if (_ccdNums.size() == 0)
	{
		errstr << "No events have been selected. Setting ONTIME values to 0..."
			   << Sas::message(Msg::App,Msg::Verbose);	
		double ontime = 0; 
		for(unsigned i=0; i<17; i++)
		{
			std::stringstream keyName;
  			keyName << "ONTIME";
  			keyName.width(2);
  			keyName.fill('0');
  			keyName << i;
  
			if( _eventTable->hasAttribute(keyName.str()) )			   			   		
	    		*_eventTable->addAttribute(keyName.str(), "Sum of GTIs for a particular CCD") 
    	  		= ontime;	
  		}
		if( _eventTable->hasAttribute("ONTIME") )			   			   		
	    	*_eventTable->addAttribute("ONTIME", "Sum of GTIs for central CCD") = ontime;	
	}

  return;
}


void XmmExposure::updateLivetime(){

  if(_instrument==Xmm::INSTRUMENT_NONE ||
	 _instrument==Xmm::RGS1 ||
	 _instrument==Xmm::RGS2) return;

  for(unsigned i=0; i<_ccdNums.size(); i++){
    std::stringstream keyName;
    keyName << "LIVETI";
    keyName.width(2);
    keyName.fill('0');
    keyName << _ccdNums[i];
    
    *_eventTable->addAttribute(keyName.str(), "Live time for a particular CCD") = _getLiveTime(i);
  }

  // Write LIVETIME keyword for the central CCD

  std::stringstream keyName;
  keyName << "LIVETI";
  keyName.width(2);
  keyName.fill('0');
  keyName << _targetCCD;
  
  if( _eventTable->hasAttribute(keyName.str()) )
    *_eventTable->addAttribute("LIVETIME", "Live time for the central CCD") 
      = _eventTable->attribute(keyName.str())->asReal();
  else if( _ccdNums.size()>0 ){
    errstr << "Live time for target "
	   << "CCD (number " << _targetCCD << ") is not available. "
	   << "Setting LIVETIME to zero."
	   << Sas::warning("NoTargetLivetime");
    *_eventTable->addAttribute("LIVETIME", "Live time for the central CCD") = 0;
  }


	// AI: SOC-SPR-2629:
	// If there are not events as a result of the filtering, the _ccdNums vector 
	// (the vector with the ccd numbers that contains events) is empty and therefore
	// the _gtiList is also empty, so the updateOntime method does not uptdate the LIVETIME
	// keyword. 
	// To fix this problem, if the _ccdNums vector is empty, we check for all the 
	// LIVETIME keywords in the header file and they are updated with 0 time. 
	if (_ccdNums.size() == 0)
	{		
		errstr << "No events have been selected. Setting LIVETIME values to 0..."
			   << Sas::message(Msg::App,Msg::Verbose);	
		double livetime = 0; 
		for(unsigned i=0; i<17; i++)
		{
			std::stringstream keyName;
  			keyName << "LIVETI";
  			keyName.width(2);
  			keyName.fill('0');
  			keyName << i;
  
			if( _eventTable->hasAttribute(keyName.str()) )			   			   		
	    		*_eventTable->addAttribute(keyName.str(), "Live time for a particular CCD") 
    	  		= livetime;	
  		}
		if( _eventTable->hasAttribute("LIVETIME") )			   			   		
	   		*_eventTable->addAttribute("LIVETIME", "Live time for the central CCD") 
      		= livetime;	
	}


  return;
}


void XmmExposure::_getGtiLists() {

  // Initialize Gti lists with tstart and tstop of EXPOSUnn extensions

  for(unsigned i=0; i<_ccdNums.size(); i++){
    std::stringstream extName;
    extName << "EXPOSU";
    extName.width(2);
    extName.fill('0');
    extName << _ccdNums[i];
 
    Block* expBlock = _set->block(extName.str());
    Gti expGti(expBlock->attribute("TSTART")->asReal(),
	       expBlock->attribute("TSTOP")->asReal());
 
    GtiList tempList;
    tempList.push_back(expGti);
    _gtiLists.push_back(tempList);
  }

  // Step thru each component of the DSS. Indetify which components have TIME filters and which
  // have CCDNR filters.  Store this information for use down stream.

  DataSubspace dss(_eventTable);
  unsigned ncomps = dss.components();
   
  if(ncomps>0){
    std::map<unsigned, DataSubspace *> ccdSubspaces;
    ccdSubspaces[0] = new DataSubspace;
    for(unsigned i=0; i<_ccdNums.size(); i++) ccdSubspaces[_ccdNums[i]] = new DataSubspace;
    
    for(unsigned icomp=0; icomp<ncomps; icomp++){
      bool noCcds = true;
      DScomp * comp = dss.component(icomp);
      for(unsigned ifilt=0; ifilt<comp->filters(); ifilt++){
	DSfilter *filter = comp->filter(ifilt);
	DSS::DSfilterComp comps = filter->comps();

	if( filter->name()=="CCDNR" ){
	  if( filter->type()==DSS::RangeT ){
    	    std::list<Range<double> > ranges = filter->ranges();
    	    std::list<Range<double> >::iterator irange;
    	    for(irange=ranges.begin(); irange!=ranges.end(); irange++){
    	      for(unsigned i=0; i<_ccdNums.size(); i++){
		if( irange->contains((double)_ccdNums[i]) )
    		  ccdSubspaces[_ccdNums[i]]->addComponent(*comp);
    	      }
    	    }
	    if( !noCcds ){
	      errstr << "The filtering component `" << comp->selectExpression() << "' contains more than "
		     << "one CCDNR filter.  As a result the exposure updating calculation may produce "
		     << "an incorrect result." << Sas::warning("ExpMultiCcdFilters");
	    }
    	    noCcds = false;
    	  }else{
    	    errstr << "The filtering expression `" << filter->selectExpression() << "' may not allow for "
    		   << "calculation of a uniform exposure for a CCD.  This filter is being ignored "
    		   << "in the exposure updating calculation."
    		   << Sas::warning("ExpBadCcdType");
    	  }
    	}else if( comps[0]=="CCDNR" || comps[1]=="CCDNR" ){
    	  errstr << "The filtering expression `" << filter->selectExpression() << "' does not allow for "
    		 << "calculation of a uniform exposure for a CCD.  This filter is being ignored "
    		 << "in the exposure updating calculation."
    		 << Sas::warning("ExpBadCcdComp");
    	}else if( filter->name()=="TIME" ){
    	  if( filter->type()!=DSS::RangeT && filter->type()!=DSS::GtiT){
    	    errstr << "The filtering expression `" << filter->selectExpression() << "' may not allow for "
    		   << "calculation of a uniform exposure for all events.  This filter is being ignored "
    		   << "in the exposure updating calculation."
    		   << Sas::warning("ExpBadTimeType");	    
    	  }
    	}else if( comps[0]=="TIME" || comps[1]=="TIME" ){
    	  errstr << "The filtering expression `" << filter->selectExpression() << "' does not allow for "
   		 << "calculation of a uniform exposure for a CCD.  This filter is being ignored "
    		 << "in the exposure updating calculation."
    		 << Sas::warning("ExpBadTimeComp");
    	}
      }
      
      if(noCcds) ccdSubspaces[0]->addComponent(*comp);
    }
    
    unsigned nccd=0;
    for(unsigned i=0; i<=_ccdNums.size(); i++){
      unsigned iccd;
      if( i==_ccdNums.size() ) iccd=0;
      else iccd = _ccdNums[i];
      DataSubspace * dss0 = ccdSubspaces[iccd];
      
      if( dss0->components() != 0 ){
	
    	// Need to check if TIME filters have been applied in a way that makes sense in terms
    	// of exposure correction, ie that different sets of events do not have different
    	// exposure corrections.
	
	// Do some special checks for global components
	if( iccd==0 && nccd!=0 ){
	  for(unsigned icomp=0; icomp<dss0->components(); icomp++){
	    for(unsigned ifilter=0; ifilter<dss0->component(icomp)->filters(); ifilter++){
	      if( dss0->component(icomp)->filter(ifilter)->name()!="TIME" ){
		errstr << "The filter expression `" << dss.selectExpression() 
		       << "' may not allow for uniform exposure correction. As a result the "
		       << "exposure updating calculation may produce an incorrect result."
		       << Sas::warning("ExpBadGlobalComp");
		break;
	      }
	    }
	  }
	}
	
	nccd++;

    	DataSubspace dssTemp = *dss0;
    	DScomp * comp0 = dssTemp.component(0);
        
    	// Check if we have just global TIME filters, or that all other filters are global.
    	
    	bool globalTime = true;
    	bool globalOther = true;
    	for(unsigned icomp=0; icomp<dss0->components(); icomp++){
   	  for(unsigned ifilter=0; ifilter<dss0->component(icomp)->filters(); ifilter++){
    	    DSfilter *filter = dss0->component(icomp)->filter(ifilter);
    	    if( !dssTemp.isGlobal(*filter) ){
    	      if(filter->name() == "TIME") globalTime = false;
    	      else globalOther = false;
    	    }
    	  }
    	}
	
    	if(globalTime){
    	  while( comp0->hasFilter("TIME", DSS::GtiT) ){
    	    DSfilter * filter0 = comp0->filter("TIME", DSS::GtiT);
    	    GtiList glist;
    	    InOGIPgtiTable oGtiTab(filter0->blockSpec());
    	    oGtiTab >> glist;
    	    if(iccd==0) for(unsigned j=0; j<_ccdNums.size(); j++) _gtiLists[j] &= glist;
    	    else _gtiLists[i] &= glist;
	    
    	    for(unsigned ifilter=0; ifilter<comp0->filters(); ifilter++)
    	      if( comp0->filter(ifilter) == filter0 ){
    		comp0->deleteFilter(ifilter);
   		break;
    	      }
    	  }
	  
    	  while( comp0->hasFilter("TIME", DSS::RangeT) ){
    	    DSfilter * filter0 = comp0->filter("TIME", DSS::RangeT);
    	    GtiList glist;
    	    std::list<Range<double> > ranges = filter0->ranges();
    	    std::list<Range<double> >::const_iterator irange;
    	    for(irange=ranges.begin(); irange!=ranges.end(); irange++){
    	      double lower, upper;
    	      RangeBound lowerB, upperB;
    	      irange->values(lower, upper, lowerB, upperB);
    	      if( lowerB==Unbounded ) lower=0;
    	      if( upperB==Unbounded ) upper=Gti::ETERNITY;
    	      Gti gti(lower, upper);
    	      glist.push_back(gti);
    	    }
	    glist.sortIt();
    	    if(iccd==0) for(unsigned j=0; j<_ccdNums.size(); j++) _gtiLists[j] &= glist;
    	    else _gtiLists[i] &= glist;
    	  
    	    for(unsigned ifilter=0; ifilter<comp0->filters(); ifilter++)
    	      if( comp0->filter(ifilter) == filter0 ){
    		comp0->deleteFilter(ifilter);
    		break;
    	      }
    	  }
	  
    	}else{
    	  // TIME filters are not all global, so check if all other filters are global.
    	  // If not, warn the user that he is on his own.
    	  if( !globalOther )
   	    errstr << "A filtering expression has been specified which may not allow for "
    		   << "calculation of a uniform exposure for all events.  Please ensure that "
    		   << "all TIME related filters are applied uniformly."
   		   << Sas::warning("ExpBadDssGlobal");
	  
   	  GtiList glist;
    	  for(unsigned icomp=0; icomp<dssTemp.components(); icomp++){
    	    GtiList compGlist(true);
    	    DScomp * comp = dssTemp.component(icomp);
    	    for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++){
    	      DSfilter *filter = comp->filter(ifilter);
    	      if(filter->name()=="TIME"){
    		GtiList filterGlist;
    		if(filter->type()==DSS::GtiT){
    		  InOGIPgtiTable oGtiTab(filter->blockSpec());
    		  oGtiTab >> filterGlist;
    		  comp->deleteFilter(ifilter);
    		}else if(filter->type()==DSS::RangeT){
    		  std::list<Range<double> > ranges = filter->ranges();
    		  std::list<Range<double> >::const_iterator irange;
    		  for(irange=ranges.begin(); irange!=ranges.end(); irange++){
    		    double lower, upper;
    		    RangeBound lowerB, upperB;
    		    irange->values(lower, upper, lowerB, upperB);
    		    if( lowerB==Unbounded ) lower=0;
    		    if( upperB==Unbounded ) upper=Gti::ETERNITY;
   		    Gti gti(lower, upper);
    		    filterGlist.push_back(gti);
    		  }
    		  comp->deleteFilter(ifilter);
    		}
    		compGlist &= filterGlist;
    	      }
    	    }
    	    glist |= compGlist;
    	  }
    	  if(iccd==0) for(unsigned j=0; j<_ccdNums.size(); j++) _gtiLists[j] &= glist;
    	  else _gtiLists[i] &= glist;
    	}
      }
    }
    for(unsigned i=0; i<_ccdNums.size(); i++) delete ccdSubspaces[_ccdNums[i]];
    delete ccdSubspaces[0];
  }//end if(ncomps>0) 
  
  for(unsigned i=0; i<_ccdNums.size(); i++){
      _gtiLists[i].reduce();
      _gtiLists[i].sortIt();
  }

  // If requested, filter the EXPOSURE extensions with their corresponding GTI.
  

  if( booleanParameter("keepfilteroutput") && booleanParameter("filterexposure") ){
    std::string tempSetName = tempnam(0,0);
    for(unsigned i=0; i<_ccdNums.size(); i++){
      std::stringstream extName;
      extName << "EXPOSU";
      extName.width(2);
      extName.fill('0');
      extName << _ccdNums[i];

      if( _gtiLists[i].size() == 0 ){
		Table * expTab = _set->table(extName.str());
		expTab->deleteRows(0, expTab->rows());
      }else{
		DataSet *tempSet = dataSetServer->open(tempSetName, DataSet::Create);
		OutOGIPgtiTable tempgti(tempSet);
		tempgti << _gtiLists[i];
		dataSetServer->close(tempSet);


	/*
	 * AI: SPR-2632. If there is a PHASE keyword in the expression, we have also to 
	 * filtering the EXPOSUnn table with the PHASE expression. So, we add it to the final
	 * expression. Right now we only allow one PHASE expression
	 * TBD: Allow the user the possibility to use more than one PHASE expression.
	 */

	std::vector<std::string> expr;

	for(unsigned icomp=0; icomp<ncomps; icomp++)
	  {	  
	    DScomp * comp = dss.component(icomp);
	    
	    for(unsigned ifilt=0; ifilt<comp->filters(); ifilt++)
	      {
		DSfilter *filter = comp->filter(ifilt);
		
		if (filter->selectExpression().find("PHASE") != std::string::npos)
		  expr.push_back(filter->selectExpression());
	      }
	  }
	
	
	if (expr.size() == 1)
	  {
	    DataFilter filter( "gti(" + tempSetName + ",TIME) &&" + expr[0] );
	    filter.doFilter(*_set->table(extName.str()));
	  }
	else
	  {
	    if(expr.size() > 1 )
	      {
		errstr << "PHASE expression to complex. Please check the LIVETIME keyword value"
		       << Sas::warning("PhaseExpressionToComplex"); 		
	      }
	    
	    DataFilter filter( "gti(" + tempSetName + ",TIME)");
	    filter.doFilter(*_set->table(extName.str()));
	  }

	unlink(tempSetName.c_str());
      }
    }
  }

  return;
}


real64 XmmExposure::_getLiveTime(unsigned ccdRef){

    unsigned ccdNr = _ccdNums[ccdRef];

    std::stringstream extName;
    extName << "EXPOSU";
    extName.width(2);
    extName.fill('0');
    extName << ccdNr;

    Table * expTab = _set->table(extName.str());

    real32 delTime = 0.;
    bool hasTimeDelCol = true;
    if( !expTab->hasColumn("TIMEDEL") ){
      hasTimeDelCol = false;
      delTime = expTab->attribute("TIMEDEL")->asReal();
    }

    ColumnData * timeCdat = expTab->column("TIME")->data();
    ColumnData * fracExpCdat = expTab->column("FRACEXP")->data();
    ColumnData * delTimeCdat = 0;
    if(hasTimeDelCol) delTimeCdat = expTab->column("TIMEDEL")->data();
    
    real64 * timeDat = timeCdat->real64Data();
    real32 * fracExpDat = fracExpCdat->real32Data();
    real32 * delTimeDat = 0;
    if(hasTimeDelCol) delTimeDat = delTimeCdat->real32Data();

    real64 timePixR = 0.5;
    if( expTab->hasAttribute("TIMEPIXR") ) 
      timePixR = expTab->attribute("TIMEPIXR")->asReal();

    GtiList gtilist = _gtiLists[ccdRef];
    const Gti * gti;

    real64 livetime = 0;
    unsigned long mRows = expTab->rows();
    for(unsigned long irow=0; irow<mRows; irow++){
      if(hasTimeDelCol) delTime = delTimeDat[irow];
      if( (gti=gtilist.inside(timeDat[irow]-delTime*timePixR)) != NULL && 
	  gti->stop() > timeDat[irow] + delTime*timePixR )
	livetime += delTime*fracExpDat[irow];
    }

    delete timeCdat;
    if(hasTimeDelCol) delete delTimeCdat;
    delete fracExpCdat;

    return livetime;
}


real64 XmmExposure::liveTime(unsigned ccdNr){
  unsigned iccdRef;
  for(iccdRef=0; iccdRef<_ccdNums.size(); iccdRef++)
    if( ccdNr == _ccdNums[iccdRef] ) break;

  if( iccdRef == _ccdNums.size() ) return 0;

  return _getLiveTime(iccdRef);
}
